package dev.fyloz.colorrecipesexplorer.config.initializers

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.config.properties.MaterialTypeProperties
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.logic.MaterialTypeLogic
import mu.KotlinLogging
import org.springframework.context.annotation.Configuration

@Configuration
@RequireDatabase
class MaterialTypeInitializer(
    private val materialTypeLogic: MaterialTypeLogic,
    private val materialTypeProperties: MaterialTypeProperties
) : AbstractInitializer() {
    private val logger = KotlinLogging.logger {}

    override fun initialize() {
        logger.debug("Executing material type initializer...")
        ensureSystemMaterialTypesExists()
        logger.debug("System material types are up to date!")
    }

    private fun ensureSystemMaterialTypesExists() {
        val systemTypes = materialTypeProperties.systemTypes.map { it.toMaterialType() }
        val oldSystemTypes = materialTypeLogic.getAll(true).toMutableSet()

        fun saveOrUpdateSystemType(type: MaterialTypeDto) {
            val storedMaterialType = materialTypeLogic.getByName(type.name)
            if (storedMaterialType != null) {
                if (!storedMaterialType.systemType) {
                    logger.info("Material type '${type.name}' already exists and will be flagged as a system type")
                    materialTypeLogic.update(storedMaterialType.copy(systemType = true))
                } else {
                    logger.debug("System material type '${type.name}' already exists")
                }
            } else {
                logger.info("System material type '${type.name}' will be created")
                materialTypeLogic.save(type)
            }
        }

        // Save new system types
        systemTypes.forEach {
            saveOrUpdateSystemType(it)
            oldSystemTypes.removeIf { c -> c.name == it.name }
        }

        // Remove old system types
        oldSystemTypes.forEach {
            logger.info("Material type '${it.name}' is not a system type anymore")
            materialTypeLogic.updateNonSystemType(it.copy(systemType = false))
        }
    }
}