package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MixDto
import dev.fyloz.colorrecipesexplorer.dtos.MixLocationDto
import dev.fyloz.colorrecipesexplorer.dtos.MixSaveDto
import dev.fyloz.colorrecipesexplorer.service.MixService
import org.springframework.context.annotation.Lazy
import org.springframework.transaction.annotation.Transactional

interface MixLogic : Logic<MixDto, MixService> {
    /** Saves the given [dto]. */
    fun save(dto: MixSaveDto): MixDto

    /** Updates the given [dto]. */
    fun update(dto: MixSaveDto): MixDto

    /** Updates the location of each mix in the given [updatedLocations]. */
    fun updateLocations(updatedLocations: Collection<MixLocationDto>)
}

@LogicComponent
class DefaultMixLogic(
    service: MixService,
    @Lazy private val recipeLogic: RecipeLogic,
    @Lazy private val materialTypeLogic: MaterialTypeLogic,
    private val mixTypeLogic: MixTypeLogic,
    private val mixQuantityLogic: MixQuantityLogic
) : BaseLogic<MixDto, MixService>(service, Constants.ModelNames.MIX), MixLogic {
    @Transactional
    override fun save(dto: MixSaveDto): MixDto {
        val recipe = recipeLogic.getById(dto.recipeId)
        val materialType = materialTypeLogic.getById(dto.materialTypeId)

        val mix = MixDto(
            recipeId = recipe.id,
            mixType = mixTypeLogic.getOrCreateForNameAndMaterialType(dto.name, materialType),
            mixQuantities = mixQuantityLogic.validateAndPrepareForMix(dto.mixQuantities)
        )

        return save(mix)
    }

    @Transactional
    override fun update(dto: MixSaveDto): MixDto {
        val materialType = materialTypeLogic.getById(dto.materialTypeId)
        val mix = getById(dto.id)

        // Update the mix type if it has been changed
        val mixType = if (mix.mixType.name != dto.name || mix.mixType.materialType.id != dto.materialTypeId) {
            mixTypeLogic.updateOrCreateForNameAndMaterialType(mix.mixType, dto.name, materialType)
        } else {
            mix.mixType
        }

        return update(
            MixDto(
                id = dto.id,
                recipeId = mix.recipeId,
                mixType = mixType,
                mixQuantities = mixQuantityLogic.validateAndPrepareForMix(dto.mixQuantities)
            )
        )
    }

    override fun updateLocations(updatedLocations: Collection<MixLocationDto>) =
        updatedLocations.forEach(::updateLocation)

    private fun updateLocation(updatedLocation: MixLocationDto) {
        service.updateLocationById(updatedLocation.mixId, updatedLocation.location)
    }
}