package dev.fyloz.colorrecipesexplorer.logic.users

import dev.fyloz.colorrecipesexplorer.SpringUserDetails
import dev.fyloz.colorrecipesexplorer.SpringUserDetailsService
import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.dtos.UserDetails
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.User
import org.springframework.context.annotation.Profile
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.stereotype.Service

interface UserDetailsLogic : SpringUserDetailsService {
    /** Loads an [User] for the given [id]. */
    fun loadUserById(id: Long, isDefaultGroupUser: Boolean = true): UserDetails
}

@Service
@RequireDatabase
class DefaultUserDetailsLogic(
    private val userLogic: UserLogic
) : UserDetailsLogic {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserById(username.toLong(), false)
        } catch (ex: NotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    override fun loadUserById(id: Long, isDefaultGroupUser: Boolean): UserDetails {
        val user = userLogic.getById(
            id,
            isSystemUser = true,
            isDefaultGroupUser = isDefaultGroupUser
        )
        return UserDetails(user)
    }
}

@Service
@Profile("emergency")
class EmergencyUserDetailsLogic(
    securityProperties: CreSecurityProperties
) : UserDetailsLogic {
    private val users: Set<UserDto>

    init {
        if (securityProperties.root == null) {
            throw NullPointerException("The root user has not been configured")
        }

        users = setOf(
            // Add root user
            with(securityProperties.root!!) {
                UserDto(
                    id = this.id,
                    firstName = "Root",
                    lastName = "User",
                    group = null,
                    password = this.password,
                    permissions = listOf(Permission.ADMIN)
                )
            }
        )
    }

    override fun loadUserByUsername(username: String): SpringUserDetails {
        return loadUserById(username.toLong(), false)
    }

    override fun loadUserById(id: Long, isDefaultGroupUser: Boolean): UserDetails {
        val user = users.firstOrNull { it.id == id }
            ?: throw UsernameNotFoundException(id.toString())

        return UserDetails(user)
    }
}
