package dev.fyloz.colorrecipesexplorer.logic

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.dtos.UserDetails
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.logic.users.DefaultJwtLogic
import dev.fyloz.colorrecipesexplorer.logic.users.jwtClaimUser
import dev.fyloz.colorrecipesexplorer.utils.base64encode
import dev.fyloz.colorrecipesexplorer.utils.isAround
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.mockk.clearAllMocks
import io.mockk.spyk
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import java.time.Instant
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class DefaultJwtLogicTest {
    private val objectMapper = jacksonObjectMapper()
    private val securityProperties = CreSecurityProperties().apply {
        jwtSecret = "XRRm7OflmFuCrOB2Xvmfsercih9DCKom"
        jwtDuration = 1000000L
    }
    private val jwtParser by lazy {
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer<Map<String, *>>(objectMapper))
            .setSigningKey(securityProperties.jwtSecret.base64encode())
            .build()
    }

    private val jwtService = spyk(DefaultJwtLogic(objectMapper, securityProperties))

    private val user = UserDto(0L, "Unit test", "User", "", null, listOf())

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun withParsedUserOutputDto(jwt: String, test: (UserDto) -> Unit) {
        val serializedUser = jwtParser.parseClaimsJws(jwt)
            .body.get(jwtClaimUser, String::class.java)

        test(objectMapper.readValue(serializedUser))
    }

    @Test
    fun buildJwt_userDetails_normalBehavior_returnsJwtStringWithValidUser() {
        val userDetails = UserDetails(user)

        val builtJwt = jwtService.buildJwt(userDetails)

        withParsedUserOutputDto(builtJwt) { parsedUser ->
            assertEquals(user, parsedUser)
        }
    }

    @Test
    fun buildJwt_user_normalBehavior_returnsJwtStringWithValidUser() {
        val builtJwt = jwtService.buildJwt(user)

        withParsedUserOutputDto(builtJwt) { parsedUser ->
            assertEquals(user, parsedUser)
        }
    }

    @Test
    fun buildJwt_user_normalBehavior_returnsJwtStringWithValidSubject() {
        val builtJwt = jwtService.buildJwt(user)
        val jwtSubject = jwtParser.parseClaimsJws(builtJwt).body.subject

        assertEquals(user.id.toString(), jwtSubject)
    }

    @Test
    fun buildJwt_user_returnsJwtWithValidExpirationDate() {
        val jwtExpectedExpirationDate = Instant.now().plusSeconds(securityProperties.jwtDuration)

        val builtJwt = jwtService.buildJwt(user)
        val jwtExpiration = jwtParser.parseClaimsJws(builtJwt)
            .body.expiration.toInstant()

        // Check if it's between 1 second
        assertTrue { jwtExpiration.isAround(jwtExpectedExpirationDate) }
    }

    // parseJwt()

    @Test
    fun parseJwt_normalBehavior_returnsExpectedUser() {
        val jwt = jwtService.buildJwt(user)
        val parsedUser = jwtService.parseJwt(jwt)

        assertEquals(user, parsedUser)
    }
}
