package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.service.MixService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test

class DefaultMixLogicTest {
    private val mixServiceMock = mockk<MixService>()
    private val recipeLogicMock = mockk<RecipeLogic>()
    private val materialTypeLogicMock = mockk<MaterialTypeLogic>()
    private val mixTypeLogicMock = mockk<MixTypeLogic>()
    private val mixQuantityLogicMock = mockk<MixQuantityLogic>()

    private val mixLogic = spyk(
        DefaultMixLogic(
            mixServiceMock,
            recipeLogicMock,
            materialTypeLogicMock,
            mixTypeLogicMock,
            mixQuantityLogicMock
        )
    )

    private val company = CompanyDto(1L, "Unit test company")
    private val recipe = RecipeDto(
        1L,
        "Unit test recipe",
        "Unit test recipe",
        "FFFFFF",
        0xf,
        null,
        null,
        false,
        "A remark",
        company,
        mutableListOf(),
        listOf()
    )
    private val materialType = MaterialTypeDto(1L, "Unit test material type", "UTMT", false)
    private val mixType =
        MixTypeDto(1L, "Unit test mix type", materialType)
    private val mixMaterial =
        MixMaterialDto(1L, MaterialDto(2L, "Unit test material", 1000f, false, materialType), 50f, 1)
    private val mixQuantities = MixQuantitiesDto(listOf(mixMaterial))
    private val mix = MixDto(recipeId = recipe.id, mixType = mixType, mixQuantities = mixQuantities)

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun setup_save_normalBehavior() {
        every { recipeLogicMock.getById(any()) } returns recipe
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { mixTypeLogicMock.getOrCreateForNameAndMaterialType(any(), any()) } returns mixType
        every { mixQuantityLogicMock.validateAndPrepareForMix(any()) } returns mixQuantities
        every { mixLogic.save(any<MixDto>()) } returnsArgument 0
    }

    private fun setup_update_normalBehavior() {
        every { recipeLogicMock.getById(any()) } returns recipe
        every { materialTypeLogicMock.getById(any()) } returns materialType
        every { mixTypeLogicMock.updateOrCreateForNameAndMaterialType(any(), any(), any()) } returns mixType
        every { mixQuantityLogicMock.validateAndPrepareForMix(any()) } returns mixQuantities
        every { mixLogic.getById(any()) } returns mix
        every { mixLogic.update(any<MixDto>()) } returnsArgument 0
    }

    @Test
    fun save_dto_normalBehavior_callsSave() {
        // Arrange
        setup_save_normalBehavior()

        val mixMaterialDto =
            MixQuantitySaveDto(mixMaterial.id, mixMaterial.material.id, mixMaterial.quantity, mixMaterial.position, false)
        val saveDto = MixSaveDto(0L, mixType.name, recipe.id, materialType.id, listOf(mixMaterialDto))

        // Act
        mixLogic.save(saveDto)

        // Assert
        verify {
            mixLogic.save(mix)
        }
    }

    @Test
    fun save_dto_normalBehavior_callsValidateAndSaveAllInMixMaterialLogic() {
        // Arrange
        setup_save_normalBehavior()

        val mixMaterialDtos =
            listOf(
                MixQuantitySaveDto(
                    mixMaterial.id,
                    mixMaterial.material.id,
                    mixMaterial.quantity,
                    mixMaterial.position,
                    false
                )
            )
        val saveDto = MixSaveDto(0L, mixType.name, recipe.id, materialType.id, mixMaterialDtos)

        // Act
        mixLogic.save(saveDto)

        // Assert
        verify {
            mixQuantityLogicMock.validateAndPrepareForMix(mixMaterialDtos)
        }
        confirmVerified(mixQuantityLogicMock)
    }

    @Test
    fun update_dto_normalBehavior_callsUpdate() {
        // Arrange
        setup_update_normalBehavior()

        val mixMaterialDto =
            MixQuantitySaveDto(mixMaterial.id, mixMaterial.material.id, mixMaterial.quantity, mixMaterial.position, false)
        val saveDto = MixSaveDto(mix.id, mixType.name, recipe.id, materialType.id, listOf(mixMaterialDto))

        // Act
        mixLogic.update(saveDto)

        // Assert
        verify {
            mixLogic.update(mix)
        }
    }

    @Test
    fun update_dto_normalBehavior_callsValidateAndSaveAllInMixMaterialLogic() {
        // Arrange
        setup_update_normalBehavior()

        val mixMaterialDtos = listOf(
            MixQuantitySaveDto(
                mixMaterial.id,
                mixMaterial.material.id,
                mixMaterial.quantity,
                mixMaterial.position,
                false
            )
        )
        val saveDto = MixSaveDto(mix.id, mixType.name, recipe.id, materialType.id, mixMaterialDtos)

        // Act
        mixLogic.update(saveDto)

        // Assert
        verify {
            mixQuantityLogicMock.validateAndPrepareForMix(mixMaterialDtos)
        }
        confirmVerified(mixQuantityLogicMock)
    }

    @Test
    fun updateLocations_normalBehavior_callsUpdateLocationByIdInServiceForEachUpdatedLocation() {
        // Arrange
        every { mixServiceMock.updateLocationById(any(), any()) } just runs

        val updatedLocations = listOf(
            MixLocationDto(1L, "A location"),
            MixLocationDto(2L, "Another location")
        )

        // Act
        mixLogic.updateLocations(updatedLocations)

        // Assert
        updatedLocations.forEach {
            verify {
                mixServiceMock.updateLocationById(it.mixId, it.location)
            }
        }
        confirmVerified(mixServiceMock)
    }
}