package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.logic.config.ConfigurationLogic
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.service.TouchUpKitService
import dev.fyloz.colorrecipesexplorer.utils.PdfDocument
import dev.fyloz.colorrecipesexplorer.utils.toByteArrayResource
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.springframework.core.io.ByteArrayResource
import kotlin.test.assertEquals

class DefaultTouchUpKitLogicTest {
    private val touchUpKitServiceMock = mockk<TouchUpKitService>()
    private val fileLogicMock = mockk<WriteableFileLogic>()
    private val configLogicMock = mockk<ConfigurationLogic>()

    private val touchUpKitLogic = spyk(DefaultTouchUpKitLogic(touchUpKitServiceMock, fileLogicMock, configLogicMock))

    private val pdfMockData = mockk<ByteArrayResource>()
    private val pdfMock = mockk<PdfDocument> {
        mockkStatic(PdfDocument::toByteArrayResource)
        mockkStatic(PdfDocument::toByteArrayResource)
        every { toByteArrayResource() } returns pdfMockData
    }

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun complete_normalBehavior_callsUpdateCompletionDateByIdInService() {
        // Arrange
        every { touchUpKitServiceMock.updateCompletionDateById(any(), any()) } just runs

        val touchUpKitId = 1L

        // Act
        touchUpKitLogic.complete(touchUpKitId)

        // Assert
        verify {
            touchUpKitServiceMock.updateCompletionDateById(touchUpKitId, any())
        }
    }

    @Test
    fun generateJobPdfResource_normalBehavior_returnsGeneratedPdf() {
        // Arrange
        every { touchUpKitLogic.generateJobPdf(any()) } returns pdfMock
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns false.toString()

        // Act
        val generatedPdfData = touchUpKitLogic.generateJobPdfResource("Unit test job")

        // Assert
        assertEquals(pdfMockData, generatedPdfData)
    }

    @Test
    fun generateJobPdfResource_normalBehavior_callsCacheJobPdf() {
        // Arrange
        every { touchUpKitLogic.generateJobPdf(any()) } returns pdfMock
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns false.toString()

        val job = "Unit test job"

        // Act
        touchUpKitLogic.generateJobPdfResource(job)

        // Assert
        verify {
            touchUpKitLogic.cacheJobPdf(job, pdfMock)
        }
    }

    @Test
    fun generateJobPdfResource_cacheEnabledAndPdfExists_returnsCachedJobPdf() {
        // Arrange
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns true.toString()
        every { fileLogicMock.exists(any()) } returns true
        every { fileLogicMock.read(any()) } returns pdfMockData

        // Act
        val pdfData = touchUpKitLogic.generateJobPdfResource("Unit test job")

        // Assert
        assertEquals(pdfMockData, pdfData)
    }

    @Test
    fun generateJobPdfResource_cacheEnabledAndPdfNotExists_generatesPdf() {
        // Arrange
        every { touchUpKitLogic.generateJobPdf(any()) } returns pdfMock
        every { touchUpKitLogic.cacheJobPdf(any(), any()) } just runs
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns true.toString()
        every { fileLogicMock.exists(any()) } returns false

        // Act
        touchUpKitLogic.generateJobPdfResource("Unit test job")

        // Assert
        verify {
            touchUpKitLogic.cacheJobPdf(any(), any())
        }
    }

    @Test
    fun cacheJobPdf_normalBehavior_callsWriteInFileLogic() {
        // Arrange
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns true.toString()
        every { fileLogicMock.write(any<ByteArrayResource>(), any(), any()) } just runs

        val job = "Unit test job"
        val pdfPath = "${Constants.FilePaths.TOUCH_UP_KITS}/$job.pdf"

        // Act
        touchUpKitLogic.cacheJobPdf(job, pdfMock)

        // Assert
        verify {
            fileLogicMock.write(pdfMockData, pdfPath, true)
        }
        confirmVerified(fileLogicMock)
    }

    @Test
    fun cacheJobPdf_cacheDisabled_doNothing() {
        // Arrange
        every { configLogicMock.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) } returns false.toString()
        every { fileLogicMock.write(any<ByteArrayResource>(), any(), any()) } just runs

        // Act
        touchUpKitLogic.cacheJobPdf("Unit test job", pdfMock)

        // Assert
        verify(exactly = 0) {
            fileLogicMock.write(any<ByteArrayResource>(), any(), any())
        }
        confirmVerified(fileLogicMock)
    }
}